from hydenv.database import HydenvDatabase
from hydenv.measurements import HydenvMeasurementsCli
from hydenv.exercises import HydenvExercises
from hydenv.backend import run as backend_run

from hydenv.scripts.hobo_import import HydenvHoboImporterCli
from hydenv.examples.examples import HydenvExamples


import fire
fire.Fire({
    'database': HydenvDatabase,
    'measurements': HydenvMeasurementsCli,
    'exercises': HydenvExercises,
    'exercise': HydenvExercises,
    'backend': backend_run,
    'hobo': HydenvHoboImporterCli,
    'example': HydenvExamples,
    'examples': HydenvExamples
})



import pandas as pd 
import progressbar
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker

from hydenv.util import env
from hydenv.models import SpaceRaw

NORMALIZE_SQL = """
-- COMPANIES
DROP TABLE IF EXISTS companies CASCADE;
CREATE TABLE companies AS
SELECT row_number() OVER (ORDER BY company_name) AS company_id, t.company_name FROM
(SELECT DISTINCT company_name FROM space_raw) t;
ALTER TABLE companies ADD CONSTRAINT pkey_companies PRIMARY KEY (company_id);
-- ROCKTES
DROP TABLE IF EXISTS rockets CASCADE;
CREATE TABLE rockets AS
SELECT row_number() OVER (ORDER BY t.rocket_name) AS rocket_id, t.rocket_name, is_active FROM
(SELECT DISTINCT
	split_part(detail, ' | ', 1) as rocket_name,
 	CASE WHEN status_rocket='StatusActive' THEN true ELSE false END as is_active
 FROM space_raw) t;
ALTER TABLE rockets ADD CONSTRAINT pkey_rockets PRIMARY KEY (rocket_id);
-- LOCATIONS
DROP TABLE IF EXISTS locations CASCADE;
CREATE TABLE locations AS
SELECT row_number() OVER (ORDER BY identifier) AS location_id, t.* FROM
(SELECT DISTINCT
 	split_part(location, ', ', 1) as identifier,
 	split_part(location, ', ', 2) as location_name,
 	CASE WHEN split_part(location, ', ', 4) = '' THEN null ELSE split_part(location, ', ', 3) END as state,
 	CASE WHEN split_part(location, ', ', 4) = '' THEN split_part(location, ', ', 3) ELSE split_part(location, ', ', 4) END AS country
 FROM space_raw) t;
ALTER TABLE locations ADD CONSTRAINT pkey_locations PRIMARY KEY (location_id);
-- SPACE
DROP TABLE IF EXISTS space CASCADE;
CREATE TABLE space AS
SELECT 
	id, datum,
	(SELECT company_id FROM companies WHERE company_name=space_raw.company_name) as company_id,
	(SELECT location_id FROM locations WHERE 
	 	identifier=split_part(location, ', ', 1) AND 
	 	location_name=split_part(location, ', ', 2)
	) as location_id,
	(SELECT rocket_id FROM rockets WHERE rocket_name=split_part(detail, ' | ', 1)) as rocket_id,
	split_part(detail, ' | ', 2) as mission_detail,
	status_mission
FROM space_raw;
ALTER TABLE space ADD CONSTRAINT pkey_space PRIMARY KEY (id);
ALTER TABLE space ADD CONSTRAINT fkey_space_location FOREIGN KEY (location_id) REFERENCES locations (location_id);
ALTER TABLE space ADD CONSTRAINT fkey_space_rocket FOREIGN KEY (rocket_id) REFERENCES rockets (rocket_id);
ALTER TABLE space ADD CONSTRAINT fkey_space_company FOREIGN KEY (company_id) REFERENCES companies (company_id);
COMMIT;
"""
DROP_SQL = """
DROP TABLE IF EXISTS companies CASCADE;
DROP TABLE IF EXISTS rockets CASCADE;
DROP TABLE IF EXISTS locations CASCADE;
DROP TABLE IF EXISTS space CASCADE;
COMMIT;
"""

class HydenvSpaceExamples:
    """
    Space Missions Example Loader.\n
    Loads Space Missions example data from Github and load it into the
    given database. The database has to be installed and initialized before. 
    Do that by hand or use the two CLI commands:
        python -m hydenv database install  --connection=postgresql://postgres:<adminpassword>@localhost:5432/postgres
        python -m hydenv database init --clean --connection=postgresql://hydenv:hydenv@localhost:5432/hydenv
    :param connection: The database URI following syntax:\n
        postgresql://<user>:<password>@<host>:<port>/<database>

    """
    def __init__(self, connection="postgresql://{usr}:{pw}@{host}:{port}/{dbname}", overwrite_url=None):
        # substitute and save
        self.__connection = env.build_connection(connection=connection)
        self.engine = create_engine(self.__connection)
        Session = sessionmaker(bind=self.engine)
        self.session = Session()

        # set some URLs
        if overwrite_url is None:
            self._data_url_or_path = 'https://raw.githubusercontent.com/data-hydenv/data/master/extra/sample/space/space.csv'
        else:
            self._data_url_or_path = overwrite_url

    def run(self, normalize=False, quiet=True):
        """
        """
        # download dataset
        df = pd.read_csv(self._data_url_or_path)

        # drop the first two columns
        df.drop(df.columns[:2], axis=1, inplace=True)

        # rename columns
        df.columns = [_.lower().replace(' ', '_')  for _ in df.columns]
        df.drop('_rocket', axis=1, inplace=True)

        # convert datum to UTC 
        df['datum'] = pd.to_datetime(df.datum, utc=True)

        # rename index to 'id'
        df.index.name = 'id'
        
        # truncate the table if it exists
        try:
            with self.engine.connect() as con:
                con.execute('TRUNCATE TABLE space_raw;')
        except:
            pass

        if not quiet:
            print('Uploading Space Missions example data...')
            bar = progressbar.ProgressBar(max_value=len(df), redirect_stdout=True)
        else:
            df.to_sql('space_raw', self.engine, if_exists='replace')

            # add the primary key
            with self.engine.connect() as con:
                con.execute('ALTER TABLE space_raw ADD CONSTRAINT pkey_space_raw PRIMARY KEY (id);')
            return

        errors = 0
        # upload missions
        for i, space in df.iterrows():
            mission = SpaceRaw(id=i, **space)

            # add
            try:
                self.session.add(mission)
                self.session.commit()
            except Exception as e:
                self.session.rollback()
                errors += 1
                if not quiet:
                    print('[%d] %s' % (i, str(e)))
            
            if not quiet:
                bar.update(i + 1)

        if errors > 10:
            if not quiet:
                print('Too many errors occured. See output above')
                return
            else:
                raise e
        
        # normalize if requested
        if normalize:
            self._normalize(quiet=quiet)
        
        if not quiet:
            print('Done.')

    def _normalize(self, quiet=True):
        if not quiet:
            print('Normalizing space schema...')
        
        # run the SQL
        with self.engine.connect() as con:
            con.execute(NORMALIZE_SQL)
    
    def drop(self):
        with self.engine.connect() as con:
            con.execute(DROP_SQL)
        


from sqlalchemy import create_engine
import pandas as pd

from hydenv.util import env


NORMALIZE_SQL = """
-- Earthquake types
drop table if exists earthquake_types CASCADE;
create table earthquake_types as 
select row_number() OVER (ORDER BY name) as type_id, name from 
(select distinct "Type" as name from earthquakes_raw) t;
alter table earthquake_types add constraint pkey_earthquake_types primary key (type_id);
-- magnitude types
drop table if exists magnitude_types CASCADE;
create table magnitude_types as 
select row_number() OVER (ORDER BY name) as type_id, name from 
(select distinct "Magnitude Type" as name from earthquakes_raw) t;
alter table magnitude_types add constraint pkey_magnitude_types primary key (type_id);
-- sources
drop table if exists earthquake_sources CASCADE;
create table earthquake_sources as
select 
	row_number() over (order by name) as source_id,
	name 
from (
select distinct name from (
	select distinct "Source" as name from earthquakes_raw
	union select distinct "Source" as name from earthquakes_raw
	union select distinct "Source" as name from earthquakes_raw
	) t_in
) t_out;
alter table earthquake_sources add constraint pkey_earthquake_sources primary key (source_id);
-- earthquakes
drop table if exists earthquakes CASCADE;
CREATE TABLE earthquakes AS
select 
	"ID" as id,
	to_timestamp("Date" || ' ' || "Time", 'MM/DD/YYYY HH24:MI') as datum,
	st_geomfromewkt('SRID=4326;POINT (' || "Longitude" || ' ' || "Latitude" || ')') as geom,
	t.type_id,
	"Magnitude" as magnitude, m.type_id as magnitude_type_id, "Magnitude Error" as magnitude_error,
	"Magnitude Seismic Stations" as magnitude_seismic_stations, 
	"Depth" as depth, "Depth Error" as depth_error, "Depth Seismic Stations" as depth_seismic_statsions,
	"Azimuthal Gap" as azimuthal_gap, "Horizontal Distance" as horizontal_distance, 
	"Horizontal Error" as horizontal_error, "Root Mean Square" as rms, 
	es.source_id, esl.source_id as location_source_id, esm.source_id as magnitude_source_id,
	case when "Status"='Reviewed' THEN true ELSE false END as is_reviewed
from earthquakes_raw r 
left outer join earthquake_types t on t.name=r."Type"
left outer join magnitude_types m on m.name=r."Magnitude Type"
left outer join earthquake_sources es on es.name=r."Source"
left outer join earthquake_sources esl on esl.name=r."Location Source"
left outer join earthquake_sources esm on esm.name=r."Magnitude Source";
ALTER TABLE earthquakes ADD CONSTRAINT pkey_earthquakes PRIMARY KEY (id);
ALTER TABLE earthquakes ADD CONSTRAINT fkey_earthquakes_type FOREIGN KEY (type_id) REFERENCES earthquake_types (type_id);
ALTER TABLE earthquakes ADD CONSTRAINT fkey_earthquakes_magnitude_types FOREIGN KEY (magnitude_type_id) REFERENCES magnitude_types (type_id);
ALTER TABLE earthquakes ADD CONSTRAINT fkey_earthquakes_sources FOREIGN KEY (source_id) REFERENCES earthquake_sources (source_id);
ALTER TABLE earthquakes ADD CONSTRAINT fkey_earthquakes_sources_location FOREIGN KEY (location_source_id) REFERENCES earthquake_sources (source_id);
ALTER TABLE earthquakes ADD CONSTRAINT fkey_earthquakes_sources_magnitude FOREIGN KEY (magnitude_source_id) REFERENCES earthquake_sources (source_id);
CREATE INDEX idx_magnitude ON earthquakes USING btree (magnitude DESC NULLS LAST);
COMMIT;
"""
DROP_SQL = """
drop table if exists earthquake_types CASCADE;
drop table if exists magnitude_types CASCADE;
drop table if exists earthquake_sources CASCADE;
drop table if exists earthquakes CASCADE;
drop table if exists earthquakes_raw CASCADE;
COMMIT;
"""


class HydenvEarthquakeExamples:
    """
    Eaarthquake Example Loader.\n
    Loads earthquake example data from Github (or other sources) and loads them into the
    given database. The database has to be installed and initialized before. 
    Do that by hand or use the two CLI commands:
        python -m hydenv database install  --connection=postgresql://postgres:<adminpassword>@localhost:5432/postgres
        python -m hydenv database init --clean --connection=postgresql://hydenv:hydenv@localhost:5432/hydenv
    :param connection: The database URI following syntax:\n
        postgresql://<user>:<password>@<host>:<port>/<database>

    """
    def __init__(self, connection="postgresql://{usr}:{pw}@{host}:{port}/{dbname}", quiet=True):
        # substitute and save
        self.__connection = env.build_connection(connection=connection)
        self.engine = create_engine(self.__connection)

        self.quiet = quiet

        # set some URLs
        self._github_url = 'https://raw.githubusercontent.com/data-hydenv/data/master/extra/spatial/earthquake/earthquakes.csv'

    def run(self, normalize=False):
        """
        """
        if not self.quiet:
            print("Earthquake example.\nDonwloading data dump...", end='')
        
        # Download data
        df = pd.read_csv(self._github_url)
        
        if not self.quiet:
            print('done.\nUploading...', end='')
        
        # Upload
        df.to_sql('earthquakes_raw', self.engine, index=None, if_exists='replace')

        if not self.quiet:
            print('done.')
        
        if normalize:
            self.__normalize()

    def __normalize(self):
        if not self.quiet:
            print('Normalizing...', end='')

        with self.engine.connect() as con:
            con.execute(NORMALIZE_SQL)
        
        if not self.quiet:
            print('done.')

    def drop(self):
        if not self.quiet:
            print('Cleaning earthquake example...', end='')
        
        with self.engine.connect() as con:
            con.execute(DROP_SQL)
        
        if not self.quiet:
            print('done.')



import os
import requests
import subprocess
from threading import Timer
import webbrowser

from hydenv.util import env

BACKEND = os.path.abspath(os.path.join(os.path.dirname(__file__), 'backend.py'))

class HydenvExercises:
    """
    Hydenv exercise tool\n
    You can use this tool to run queries against a connected database instance.
    It is also possible to compare your result to prepared solutions.
    :param connection: The database URI following syntax:\n
        postgresql://<user>:<password>@<host>:<port>/<database>
    :param backend: The backend server to process the SQL.\n
        defaults to 'local', to run on localhost. Can be overwritten by a
        remote API server.
    """
    def __init__(self, connection="postgresql://{usr}:{pw}@{host}:{port}/{dbname}", backend='local'):
        self.connection = env.build_connection(connection=connection)

        # backend settings
        self.backend_use_debug = False
        self.allow_external = False     
        self._backend_proc = None
        self._backend_thread = None

        if backend == 'local':
            self.backend = 'http://localhost:5000/'
            self.backend_use_debug = True
        elif backend == 'extern' or backend == 'external':
            self.backend = 'http://localhost:5000/'
            self.backend_use_debug = False
            self.allow_external = True
        else:
            self.backend = backend
        print(self.backend)

    def start(self, no_browser=False):
        """
        Start the Exercises Application.\n
        This runs a backend server to solve SQL exercises in the 
        connected database. It will automatically start a browser and 
        navigate to the application. If the browser does not open, you can 
        find the app at: http://localhost:5000
        :param connection: The database URI following syntax:\n
            postgresql://<user>:<password>@<host>:<port>/<database>
        :param no_browser: If flag is set, the application will start supress
            the webbrowser startup.
        """
        print(self.backend)
        try:
            res = requests.get(self.backend + 'api/v1/ping')
            if res.status_code != 200:
                raise requests.ConnectionError()
            print('Backend instance already running')
        except requests.ConnectionError:
            self._run_backend_server(no_browser=no_browser)

    def _run_backend_server(self, no_browser=False):
        """
        On local machines, we need to start the functions-framework in a 
        subprocess.
        """
        print('Starting backend server: %s' % self.backend)
        
        # add the database connection to evironment variables

        # build command
        cmds = ['python', BACKEND, '--db_uri=%s' % self.connection]
        if self.backend_use_debug:
            cmds.append('--debug')
        if self.allow_external:
            cmds.extend(['--host', '0.0.0.0'])
#            subprocess.call(['iptables', '-t', 'nat', '-A', 'PREROUTING', '-p', 'tcp', '--dport', '80', '-j', 'REDIRECT', '--to-port', '8080'])
        
        # execute
        print('Backend is running.')
        if not no_browser:
            t = Timer(0.8, webbrowser.open_new_tab, args=[self.backend + 'index.html'])
            t.start()
        self._backend_proc = subprocess.call(cmds)
        

    def run_sql(self, sql: str, safe=True):
        """
        Run SQL code against the given connection and return the result.
        :param sql: The sql code to be executed
        :param safe: Only allow SELECT statements
        """
        if safe and not sql.lower().startswith('select'):
            print('The SQL statement is not allowed.')
            return
        
        res = requests.get(self.backend + 'run', params={'sql': sql})
        if res.status_code == 200:
            return res.json()
        else:
            print('An error occured.\nPlease report to the developer: mirko@hydrocode.de')
            return res.json()




import decimal
import datetime


def sqlencoder(data):
    if isinstance(data, datetime.date):
        return data.isoformat()
    if isinstance(data, decimal.Decimal):
        return float(data)
    return data


import os
import sys
import json
from datetime import datetime as dt

from flask import Flask, jsonify, request, send_from_directory
from flask_cors import CORS
from sqlalchemy import create_engine
from sqlalchemy.exc import ResourceClosedError

from hydenv.util.json import sqlencoder
from hydenv.database import HydenvDatabase
from hydenv import __version__ as hydenv_version

app = Flask(__name__)
CORS(app)


@app.route('/<path:path>', methods=['GET'])
def static_files(path: str):
    if path.endswith('js'):
        return send_from_directory('./exercises-js/', path, mimetype='application/javascript')
    return send_from_directory('./exercises-js/', path)

@app.route('/')
def index():
    return send_from_directory('./', 'index.html')

@app.route('/api/v1/ping', methods=['GET'])
def ping():
    return jsonify({
        'message': 'Raw data is awesome',
        'version': {
            'python': '%d.%d.%d' % sys.version_info[:3],
            'python_text': sys.version,
            'hydenv': hydenv_version
        }
    }), 200


@app.route('/api/v1/execute', methods=['GET'])
def execute():
    t1 = dt.now()
    # get passed data
    data = request.json if request.json else {}
    data.update(request.args)

    # get SQL query
    if 'safe' in data and data['safe'].lower() in ('f', 'false'):
        safe = False
    else:
        safe = True
    sql = data.get('sql')
    explain = data.get('explain', False)
    if sql is None:
        return jsonify({'message': 'No SQL query given.'}), 404
    else:
        sql = sql.replace('%', '%%')
    
    # build the CLI
    cli = HydenvDatabase(connection=app.config['DB_URI'])

    # run
    t2 = dt.now()
    try:
        res = cli.execute(sql=sql, safe=safe, json=True)
    except ResourceClosedError:
        res = [
            dict(message="This query did not return any result rows.\nMost likely you ran an INSERT, UPDATE or CREATE. If not, check if this was an error.")
            ]
    except Exception as e:
        return jsonify({
                'message': 'Run errored!',
                'params': data,
                'error': True,
                'data': [dict(errorMessage=str(e))],
                'perf': dict(startupTime=(t2 - t1).total_seconds())
            }, )

    # extract the data
    try:
        result_data = [{k: sqlencoder(v) for k,v in r.items()} for r in res]
    except Exception as e:
        result_data = [dict(message="Error parsing the result data.\n%s" % str(e))]

    t3 = dt.now()
    
    # build response object
    response = {
        'message': 'Run successful',
        'params': data,
        'data': result_data,
        'perf': dict(
            backendTime=(t3 - t2).total_seconds(), 
            executionTime=(t3 - t1).total_seconds(),
            startupTime=(t2 - t1).total_seconds(),
        )
    }

    # run explain if needed
    if explain:
        # get the format
        if explain not in ['text', 'json', 'xml', 'yaml']:
            fmt = 'text'
        else:
            fmt = explain
        try:
            exp_text = cli.explain(sql=sql, fmt=fmt)
        except Exception as e:
            exp_text = str(e)
        response['explain'] = exp_text
    
    # return
    return jsonify(response)


@app.route('/api/v1/explain', methods=['GET'])
def explain():
    data = request.json if request.json else {}
    data.update(request.args)

    # get SQL query and params
    full = data.get('full', False)
    fmt = data.get('format', data.get('fmt', 'json'))
    sql = data.get('sql')
    if sql is None:
        return jsonify({'message': 'No SQL query given.'}), 404
    
    # run the CLI
    cli = HydenvDatabase(connection=app.config['DB_URI'])
    exp = cli.explain(sql=sql, fmt=fmt, full=full)

    # return
    if fmt == 'json':
        return jsonify(exp)
    else:
        return exp




def run(db_uri, debug=False, port=5000, host='localhost'):
    """
    Hydenv Exercise backend server\n
    For local execution of exercises. Usually, the exercise 
    CLI starts the backend and you don't need to start it 
    manually. However, you have more control over the server, but 
    you need to start it before you start the exercises.
    """
    app.config['DB_URI'] = db_uri
    app.run(debug=debug, port=port, host=host)


if __name__=='__main__':
    import fire
    fire.Fire(run)


import os
import json

NAME_MAP = dict(
    usr='POSTGRES_USER',
    pw='POSTGRES_PASSWORD',
    host='POSTGRES_HOST',
    port='POSTGRES_PORT',
    dbname='POSTGRES_DBNAME'
)

def store_file(**kwargs):
    fname = os.path.abspath(os.path.join(os.path.expanduser('~'), 'hydenv.env'))
    config = read_file()
    # update
    for k,v in kwargs.items():
        config[NAME_MAP[k]] = v
    
    with open(fname, 'w') as f:
        json.dump(config, f)

def read_file() -> dict:
    fname = os.path.abspath(os.path.join(os.path.expanduser('~'), 'hydenv.env'))
    if os.path.exists(fname):
        with open(fname, 'r') as f:
            config = json.load(f)
    else:
        config = dict()
    return config

def expose(**kwargs):
    for k,v in kwargs.items():
        os.environ[k] = v


def build_connection(connection="postgresql://{usr}:{pw}@{host}:{port}/{dbname}") -> str:
    # check for replacements
    fargs = read_file()
    args = dict()
    if '{usr}' in connection:
        args['usr'] = os.environ.get('POSTGRES_USER', fargs.get('POSTGRES_USER', 'hydenv'))
    if '{pw}' in connection:
        args['pw'] = os.environ.get('POSTGRES_PASSWORD', fargs.get('POSTGRES_PASSWORD', 'hydenv'))
    if '{host}' in connection:
        args['host'] = os.environ.get('POSTGRES_HOST', fargs.get('POSTGRES_HOST', 'localhost'))
    if '{port}' in connection:
        args['port'] = os.environ.get('POSTGRES_PORT', fargs.get('POSTGRES_PORT', 5432))
    if '{dbname}' in connection:
        args['dbname'] = os.environ.get('POSTGRES_DBNAME', fargs.get('POSTGRES_DBNAME', 'hydenv'))
        
    # substitute and return
    return connection.format(**args)





from progressbar import ProgressBar

from hydenv.examples.hobo import HydenvHoboExamples
from hydenv.examples.space import HydenvSpaceExamples
from hydenv.examples.customers import HydenvCustomerExamples
from hydenv.examples.osm import HydenvOSMExamples
from hydenv.examples.earthquakes import HydenvEarthquakeExamples
from hydenv.examples.wbd import HydenvWorldBorderExample

CLIs = [HydenvHoboExamples, HydenvSpaceExamples, HydenvCustomerExamples, HydenvOSMExamples, HydenvEarthquakeExamples]

class HydenvExamples:
    r"""
    Example Loader.\n
    Loads example files from Github (or other sources) and loads them into the
    given database. The database has to be installed and initialized before. 
    Do that by hand or use the two CLI commands:
        python -m hydenv database install  --connection=postgresql://postgres:<adminpassword>@localhost:5432/postgres
        python -m hydenv database init --clean --connection=postgresql://hydenv:hydenv@localhost:5432/hydenv
    :param connection: The database URI following syntax:\n
        postgresql://<user>:<password>@<host>:<port>/<database>
    """
    def __init__(self, connection="postgresql://{usr}:{pw}@{host}:{port}/{dbname}", quiet=False):
        # substitute and save
        self.__connection = connection
        self.quiet = quiet

    def hobo(self, terms='all', only=None, data_path='download'):
        """
        Import HOBO data.\n
        This high level script downloads all neccessary resources for the HOBO 
        measureing campaigns and uploads them to the specified database.
        You can limit the metadata by terms (like WT17, WT18 etc.) and upload only
        'metadata' or 'data'. Default is to upload both.
        :param terms: Either 'all' (default) or a specific term short name (like WT18) to use
        :param data_path: Either 'download' (default) or a local path to the data
        :param only: If set, only the given entities will be uploaded.
        """
        cli = HydenvHoboExamples(connection=self.__connection)
        return cli.run(terms=terms, data_path=data_path, only=only, quiet=self.quiet)

    def space(self, normalize=False):
        """
        Import Space Missions example data.\n
        This high level script downloads a database of all space Mission from 
        1957 until early 2020. The raw, de-normalized dataset is installed along 
        with the hydenv database and used in lecture to introduce SQL and normalization.
        The final normalized structure can be added using the normalize flag.
        :param normalize: If set, the data will be normalized before uplaod.
        """
        cli = HydenvSpaceExamples(connection=self.__connection)
        return cli.run(normalize=normalize, quiet=self.quiet)

    def customers(self, k=500, normalize=False):
        """
        Import Customers example data.\n
        This high level script create a set of k random fake customer purchases 
        in fake stores. The data is not normalized and can be used for the 
        normalization sessions.
        :param k: number of purchases to create
        :param normalize: If True, the normalized scheme will be added as well.
        """
        cli = HydenvCustomerExamples(connection=self.__connection)
        return cli.run(quiet=self.quiet, k=k, normalize=normalize)

    def customer(self, **kwargs):
        return self.customers(**kwargs)

    def osm(self, action, *args, **kwargs):
        """
        Import OSM example data.\n
        This high level script downloads data from the OSM database and imports
        it into the hydenv database at given connection. You have to specify, 
        which example set should be loaded using the `action` parameter.
        There are the following example sets available:
        
        * energiewende
        * city-districts

        Additionally, you can use the more low level endpoints to access any data.

        * nodes (for POINT geometries)
        * ways (for LINESTRING and POLYGON geometries)

        Using the low level endpoints, you need to specify the `boundary`,
        `attribute` and `value` parameters.
        :param action: The CLI action to perfom. See docs above.
        :param boundary: string - Any valid OSM administrative boundary to limit
            the results to, like `"Baden-Württemberg"` or `"Freiburg im Breisghau"`.
        :param attribute: string - Any valid OSM Tag attribute for filtering.
            In most cases, this will be `amenity`.
        :param value: string - filter value. This is the value that the 
            attribute should have. If value is not given, all nodes with having 
            the tag will be returned. That can be a lot.
        """
        cli = HydenvOSMExamples(connection=self.__connection)

        if action.lower() == 'energiewende':
            return cli.energiewende(quiet=self.quiet, **kwargs)
        elif action.lower() == 'city-districts' or action.lower() == 'city_districts':
            return cli.city_districts(quiet=self.quiet, **kwargs)
        elif action.lower() == 'counties':
            return cli.counties(quiet=self.quiet, **kwargs)
        elif action.lower() == 'nodes':
            return cli.run('nodes', *args, quiet=self.quiet, **kwargs)
        elif action.lower() == 'way' or action.lower() == 'ways':
            return cli.run('way', *args, quiet=self.quiet, **kwargs)
        else:
            print("'%s' is not a supported OSM example." % action)

    def earthquake(self, normalize=False):
        """
        Import Earthquakes example data.\n
        This high level script downloads a database of earthquakes from 
        1967 to 2016 with a magnitude > 5.5. This dataset can be used for an advanced
        normalization exercise and as a spatial dataset with global coverage.
        The final normalized structure can be added using the normalize flag.
        :param normalize: If set, the data will be normalized before uplaod.
        """
        cli = HydenvEarthquakeExamples(connection=self.__connection, quiet=self.quiet)
        return cli.run(normalize=normalize)

    def world_borders(self):
        """
        Load a minified version of the World Border dataset.\n
        This dataset should only be used as a overview for maps. There is 
        only very limited information in the tables and the border geometries
        are very coarse.
        """
        cli = HydenvWorldBorderExample(connection=self.__connection)
        return cli.run(quiet=self.quiet)

    def clean(self): 
        """
        Cleaning Up.\n
        This command will clean all additional tables that were created by the CLI 
        which do not have a model representation in `hydenv.models`. This includes
        the normalized examples of space and customers.
        Usually, this is only called by the `database --clean` command and a 
        manual invocation is not needed
        """
        if not self.quiet:
            bar = ProgressBar(max_value=len(CLIs), redirect_stdout=True)
        
        # drop all extra tables
        for i, CLI in enumerate(CLIs):
            if hasattr(CLI, 'drop'):
                cli = CLI(connection=self.__connection)
                cli.drop()

                if not self.quiet:
                    bar.update(i + 1)


if __name__=='__main__':
    import fire
    fire.Fire(HydenvExamples)
